// -*- c-basic-offset: 4; indent-tabs-mode: nil -*-

/* ***** BEGIN LICENSE BLOCK *****
Version: MPL 1.1/GPL 2.0/LGPL 2.1

The contents of this file are subject to the Mozilla Public License
Version 1.1 (the "License"); you may not use this file except in
compliance with the License. You may obtain a copy of the License at
http://www.mozilla.org/MPL/

Software distributed under the License is distributed on an "AS IS"
basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
License for the specific language governing rights and limitations
under the License.

The Original Code is Standard-Sitemap Navigator code.

The Initial Developer of the Original Code is the
Standard-Sitemap Team.

Contributor(s): (see license.txt)

Alternatively, the contents of this file may be used under the terms of
either the GNU General Public License Version 2 or later (the "GPL"), or
the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
in which case the provisions of the GPL or the LGPL are applicable instead
of those above. If you wish to allow use of your version of this file only
under the terms of either the GPL or the LGPL, and not to allow others to
use your version of this file under the terms of the MPL, indicate your
decision by deleting the provisions above and replace them with the notice
and other provisions required by the GPL or the LGPL. If you do not delete
the provisions above, a recipient may use your version of this file under
the terms of any one of the MPL, the GPL or the LGPL.
***** END LICENSE BLOCK ***** */

// This is loaded with every window.  It keeps track of all events
// pertaining to which sitemap is associated with which tab, and which
// page is currently displayed (so a particular sitemap-rendering
// component can tell how to highlight it).

Components.utils.import("resource://modules.standard-sitemap.org/sslib.jsm");
Components.utils.import("resource://modules.standard-sitemap.org/service.jsm");


window["org.standard-sitemap.monitor"] = {
    lib : Application.storage.get("org.standard-sitemap.library", null),

    oSitemapService : null,

    TabAgent : function(oSitemapService, oBrowser) {
        this.oBrowser = oBrowser;
        this.oLastLocation = this.oBrowser.currentURI;
        if (this.oLastLocation)
            this.oLastLocation = this.oLastLocation.clone();
        this.oSitemapService = oSitemapService;
        this.aListeners = [ ];
        this.mDisplayLevel = { };
    },


    // Get the sitemap service instance.  Ensure it exists if not
    // already.
    getService : function() {
        if (!this.oSitemapService) {
            this.oSitemapService =
                Application.storage.get("org.standard-sitemap.service", null);
        }

        return this.oSitemapService;
    },

    // Get the TabAgent for a specific tab.  It will be
    // created and attached to the tab's browser if it does not
    // already exist.
    getAgentForTab : function(tab) {
        if (!tab)
            return null;
        var browser = getBrowser().getBrowserForTab(tab);
        return this.getAgentForBrowser(browser);
    },

    // Get the TabAgent for a specific browser.  It
    // will be created and attached to the browser if it does not
    // already exist.
    getAgentForBrowser : function(browser) {
        var result = browser.getUserData('org.standard-sitemap.tabagent');
        if (result == null) {
            result = new (this.TabAgent)(this.getService(), browser);
            browser.setUserData('org.standard-sitemap.tabagent', result, null);
        }
        return result;
    },

    // Private.  Call this when a tab is being closed.
    deregisterTab : function(tab) {
        var browser = getBrowser().getBrowserForTab(tab);
        if (!browser) {
            this.lib.debug('No browser');
            return;
        }

        var agent = this.getAgentForBrowser(browser);
        agent.applyStyles(null);

        browser.removeProgressListener(agent.progressListener);

        // TODO: Remove other listeners?
    },

    // Private.  Call this when a new tab has been created.
    registerTab : function(tab) {
        // Which browser runs this tab?
        var browser = getBrowser().getBrowserForTab(tab);
        if (!browser) {
            this.lib.debug('No browser');
            return;
        }
        if (false)
            this.lib.debug('New tab ' + browser.parentNode.id);

        // Make sure that the browser for this tab has an agent.
        var agent = this.getAgentForBrowser(browser);

        var location = browser.currentURI;
        if (location)
            agent.recordLocationChange(location, "register");

        var monitor = this;

        // Watch this tab for change of location.
        agent.progressListener = {
            QueryInterface : function(aIID) {
                if (aIID.equals(Components.interfaces
                                .nsIWebProgressListener) ||
                    aIID.equals(Components.interfaces
                                .nsISupportsWeakReference) ||
                    aIID.equals(Components.interfaces.nsISupports))
                    return this;
                throw Components.results.NS_NOINTERFACE;
            },

            onStateChange : function() { },
            onProgressChange : function() { },
            onStatusChange : function() { },
            onSecurityChange : function() { },

            onLocationChange : function(webProgress,
                                        request, location) {
                agent.recordLocationChange(location);
            }
        };
        browser.addProgressListener(agent.progressListener);

        // Watch out for content being loaded into this tab.  This
        // could mean that the current sitemap URI has changed.
        var bIgnore = false;
        var onShowOrLoad = function(event) {
            var document = event.target;

            var eventBrowser = getBrowser().getBrowserForDocument(document);
            if (eventBrowser != browser)
                return;

            // Don't do anything for subframes.
            if (!document.defaultView.top ||
                document.location != document.defaultView.top.location)
                return;

            // If we get DOMContentLoaded, we don't need to handle the
            // subsequent pageshow.
            if (event.type == 'DOMContentLoaded')
                bIgnore = true;
            else if (event.type == 'pageshow' && bIgnore) {
                bIgnore = false;
                return;
            }

            if (true)
                monitor.lib.debug('Document loaded: (' + event.type + ') ' +
                                  browser.currentURI.spec);
            agent.bNewPage = false;

            // Ensure that the agent holds the correct sitemap URI.
            monitor.checkForSitemap(browser, agent);
        };
        browser.addEventListener('DOMContentLoaded', onShowOrLoad, false);
        browser.addEventListener('pageshow', onShowOrLoad, false);

        // Check now for a sitemap - content might already have been
        // loaded.
        this.checkForSitemap(browser, agent);
    },

    checkForSitemap : function(browser, agent) {
        var sSource;
        var oSitemapURI = null;

        // If the sitemap URI is unknown, try to get it from the HTTP
        // header, using the document's URI as a base.
        if (!oSitemapURI) {
            try {
                oSitemapURI = this.getSitemapURIFromHTTP(browser);
                if (oSitemapURI)
                    sSource = 'HTTP';
            } catch (ex) {
                // Do nothing.
            }
        }

        // If the sitemap URI is unknown, try to get it from the
        // <head> element of the document, using <base href> as the
        // base if set, or the document's URI otherwise.
        if (!oSitemapURI) {
            oSitemapURI = this.getSitemapURIFromHTML(browser);
            if (oSitemapURI)
                sSource = 'HTML';
        }

        // Otherwise, if we're looking at a sitemap document, and
        // we're in webmaster mode, let's display this sitemap.
        if (!oSitemapURI && this.lib.getBoolPref('modeWebmaster')) {
            oSitemapURI = this.getSitemapURIFromLocation(browser);
            if (oSitemapURI)
                sSource = 'Sitemap';
        }

        this.lib.debug('Subject <' + browser.currentURI.spec + '>: sitemap <' +
                       (oSitemapURI ? oSitemapURI.spec : null) +
                       '>' + (sSource ? ' from ' + sSource : ''));

        // Get the load flags to detect whether we should reload the
        // sitemap.
        var oDocShell = browser.docShell;
        if (oDocShell.loadType &
            Components.interfaces.nsIDocShell.LOAD_CMD_RELOAD) {
            this.lib.debug('Reload detected');
            agent.invalidateSitemap(true, true);
        }

        agent.recordSitemapChange(oSitemapURI);
    },

    getSitemapURIFromLocation : function(browser) {
        var oDocument = browser.contentDocument;
        if (oDocument.contentType != 'text/xml' &&
            oDocument.contentType != 'application/xml')
            return null;
        if (oDocument.documentElement.localName != 'sitemap')
            return null;
        if (!this.lib.issmns(oDocument.documentElement.namespaceURI))
            return null;
        return browser.currentURI;
    },

    // Get the X-Standard-Sitemap-Location HTTP header which came with
    // the current document.
    getSitemapURIFromHTTP : function(browser) {
        var oDocShell = browser.docShell;
        var oHTTPChannel = oDocShell.currentDocumentChannel
            .QueryInterface(Components.interfaces.nsIHttpChannel);

        // First try using the namespaced extension headers (RFC2774).
        var lib = this.lib;
        var uristr = '"' + lib.sitemapNS[0] + '"';
        var urilen = uristr.length;
        var visitor = {
            headers : [ ],
            exp : /^;\s*[nN][sS]\s*=\s*([0-9][0-9])/,
            visitHeader : function(name, val) {
                if (name.toLowerCase() != "man" &&
                    name.toLowerCase() != "opt")
                    return;
                //lib.debug('Found [' + val + "] == " + uristr);
                var found = false;
                if (!lib.stringStartsWith(val, uristr))
                    return;
                var parts = this.exp.exec(val.substring(urilen));
                if (parts.length < 2)
                    return;
                //lib.debug('Pushing ' + parts[1] + '-Location');
                this.headers.push(parts[1] + '-Location');
            },
        };
        oHTTPChannel.visitResponseHeaders(visitor);

        // Fall back to the experimental namespace header.
        visitor.headers.push("X-Standard-Sitemap-Location");

        for each (var hdr in visitor.headers) {
            //this.lib.debug('Testing hdr ' + hdr);
            var val = oHTTPChannel.getResponseHeader(hdr);
            if (!val) continue;
            return this.lib.makeURI(browser.currentURI.resolve(val));
        }
        return null;

        var sSitemapURI =
            oHTTPChannel.getResponseHeader("X-Standard-Sitemap-Location");
        if (!sSitemapURI)
            return null;
        return this.lib.makeURI(browser.currentURI.resolve(sSitemapURI));
    },

    // Read <link rel="standard-sitemap"> @href and resolve it.
    getSitemapURIFromHTML : function(browser) {
        var oDocument = browser.contentDocument;
        var linkTags = oDocument.getElementsByTagName('link');

        // Build up a namespace table first.
        var prefixes = [];
        prefixes.push('standard-sitemap');
        prefixes.push('sitemap');
        for (var i = 0; i < linkTags.length; i++) {
            var rel = linkTags[i].getAttribute('rel');
            if (!rel)
                continue;
            var parts = rel.toString().trim().split(/\s+/);
            for each (var part in parts) {
                part = part.toLowerCase();
                if (!this.lib.stringStartsWith(part, 'schema.'))
                    continue;
                if (!this.lib.issmns(linkTags[i].getAttribute('href')))
                    continue;
                prefixes.push(part.substring(7) + '.location');
            }
        }

        for (var i = 0; i < linkTags.length; i++) {
            var rel = linkTags[i].getAttribute('rel');
            if (!rel)
                continue;

            var parts = rel.toString().trim().split(/\s+/);
            for each (var part in parts) {
                part = part.toLowerCase();
                var failed = true;
                for each (var prefix in prefixes) {
                    if (part == prefix) {
                        failed = false;
                        break;
                    }
                }
                if (failed) continue;

                var sSitemapURI = linkTags[i].getAttribute('href');
                if (!sSitemapURI)
                    continue;

                // Get the base to resolve the address against.
                var oBaseURI = browser.currentURI;
                var baseTags = oDocument.getElementsByTagName('base');
                if (baseTags.length != 0) {
                    var baseHRef = baseTags[0].getAttribute('href');
                    if (baseHRef)
                        oBaseURI =
                        this.lib.makeURI(oBaseURI.resolve(baseHRef));
                }

                // Now resolve the address.
                return this.lib.makeURI(oBaseURI.resolve(sSitemapURI));
            }
        }
        return null;
    },

    lastSelectedAgent : null,

    reportTabSelect : function(agent) {
        if (agent == this.lastSelectedAgent)
            return;
        if (this.lastSelectedAgent)
            this.lastSelectedAgent.reportTabVisibility(false);
        if (agent)
            agent.reportTabVisibility(true);
        this.lastSelectedAgent = agent;
    },

    // Make sure we can hook ourselves into any tab that is created.
    start : function() {
        var tabs = getBrowser().tabContainer;
        if (!tabs) return;

        // Make sure we register all existing tabs.
        for (var i = 0; i < tabs.itemCount; i++)
            this.registerTab(tabs.getItemAtIndex(i));
        if (false)
            this.lib.debug('Registered ' + tabs.itemCount + ' existing tabs');

        var self = this;

        // Make sure we detect other tabs being created.
        tabs.addEventListener('TabOpen', function(event) {
            self.registerTab(event.target);
        }, false);
        if (false)
            this.lib.debug('Ready for new tabs');

        // Detect tabs being destroyed too.
        tabs.addEventListener('TabClose', function(event) {
            self.deregisterTab(event.target);
        }, false);

        // Watch out for tabs in this window being switched, and keep
        // the corresponding agents up-to-date.  We only need this to
        // apply styles using the StylesheetService, so we should be
        // able to discard this function if we can get Media Queries
        // working in our favour.
        tabs.addEventListener('TabSelect', function(event) {
            var browser = gBrowser.selectedTab.linkedBrowser;
            var agent = self.getAgentForBrowser(browser);
            self.reportTabSelect(agent);
        }, false);

        // Make sure the agent of the currently selected tab knows
        // that it is selected.
        this.reportTabSelect(this.getAgentForTab(gBrowser.selectedTab));
    },

    end : function() {
        var tabs = getBrowser().tabContainer;
        for (var i = 0; i < tabs.itemCount; i++)
            this.deregisterTab(tabs.getItemAtIndex(i));
    },
};


// Each agent acts on behalf of a single tab.  It can be obtained by
// calling:
//
// window.window["org.standard-sitemap.monitor"].getAgentForTab/Browser(tab/browser).
//
// Clients (e.g. a sitemap sidebar) should use agent.getLocation(),
// agent.getSitemapURI() and agent.getSitemap() to get current
// sitemap-related information for the tab represented by 'agent'.
// They can also call agent.addListener(func) to detect when anything
// changes.
window["org.standard-sitemap.monitor"].TabAgent.prototype = {
    lib : Application.storage.get("org.standard-sitemap.library", null),

    oSitemapService : null,

    oSitemapURI : null,

    oSitemap : null,

    bReloadNext : false,

    bSkipCacheNext : false,

    bNewPage : true,

    aListeners : [ ],

    oBrowser : null,

    bVisible : false,

    oLastLocation : null,

    mDisplayLevel : { },
    iLastDisplayLevel : 0,

    getLocation : function() {
        return this.oLastLocation;
    },

    getSitemapURI : function() {
        return this.oSitemapURI;
    },

    // Get the StandardSitemap object for the represented tab.
    // fCallback will be passed the sitemap as its first argument.
    getSitemap : function(fCallback) {
        var self = this;

        if (this.oSitemap != null) {
            var sm = this.oSitemap;
            if (fCallback)
                setTimeout(function() { fCallback(sm); }, 0);
            return;
        }

        if (this.oSitemapURI == null) {
            if (fCallback)
                setTimeout(function() {
                    if (false)
                        self.lib.debug('URI null, so sitemap null');
                    fCallback(null);
                }, 0);
            return;
        }

        var self = this;
        var wrappedCallback = function(oSitemap) {
            if (false)
                self.lib.debug('got sitemap, sending to client' +
                               (oSitemap ? '' : ' (null)'));
            self.lib.debug('Sitemap languages: ' + oSitemap.GetLanguages());
            self.oSitemap = oSitemap;
            self.updateStyles(oSitemap);
            if (fCallback)
                fCallback(oSitemap);
        };

        this.oSitemapService.LoadSitemapInternal(window,
                                                 this.oSitemapURI.spec,
                                                 wrappedCallback,
                                                 this.bReloadNext,
                                                 this.bSkipCacheNext);

        this.bReloadNext = false;
        this.bSkipCacheNext = false;
    },

    // Add a listener to be informed whenever the sitemap state of
    // this tab changes.  The argument should be a function.
    addListener : function(fListener) {
        for each (var el in this.aListeners) {
            if (el == fListener)
                return;
        }

        this.aListeners.push(fListener);
    },

    removeListener : function(fListener) {
        for (var i in this.aListeners) {
            if (this.aListeners[i] == fListener) {
                this.aListeners.splice(i, 1);
                return;
            }
        }
    },

    // Discard the current sitemap data.  A client might call this if
    // it needs to force a reload of a sitemap (e.g. the sidebar's
    // reload button is clicked).  The two parameters will be
    // recorded, and later passed to
    // StandardSitemapService#LoadSitemapInternal.
    invalidateSitemap : function(bReloadNext, bSkipCacheNext) {
        this.oSitemap = null;
        this.bReloadNext = bReloadNext;
        this.bSkipCacheNext = bSkipCacheNext;
    },

    // Tell all clients of this agent that the document location has
    // changed, including movements within the URI as the anchor
    // reference changes.  This should only be called from this
    // object.
    recordLocationChange : function(oNewLocation, sSuffix) {
//         this.lib.debug('locationChange(' +
//                     (oNewLocation ? oNewLocation.spec : null) + ', ' +
//                     sSuffix + ')');
        this.bNewPage = true;
        this.oLastLocation = oNewLocation.clone();
        this.informListeners('location-change' +
                             (sSuffix ? '-' + sSuffix : ''));
        this.updateStyles();
    },

    // Indicate that the sitemap URI of the current page has been
    // identified.  'null' means that no sitemap was specified.  If
    // the new address is the same as the old, nothing happens.
    // Otherwise, the new address is recorded, any currently cached
    // sitemap data is discarded, and all the agent's clients will be
    // informed.  This should only be called from this object.
    recordSitemapChange : function(oNewSitemapURI) {
        if (false)
            this.lib.debug('sitemapChange(' +
                           (oNewSitemapURI ? oNewSitemapURI.spec: null) + ')');

        if (!this.bReloadNext) {
            // Has the sitemap URI changed?
            if (!this.oSitemapURI && !oNewSitemapURI)
                // Both old and new are null.
                return;
            if (this.oSitemapURI && oNewSitemapURI &&
                this.oSitemapURI.equals(oNewSitemapURI)) {
                // The URI has not changed - let the users know that
                // we now know we're not changing, so they don't have
                // to wait forever for 'sitemap-change'.
                this.informListeners('in-site-change');
                return;
            }
        }

        // Record the new state.
        this.oSitemapURI = oNewSitemapURI;
        this.oSitemap = null;

        this.informListeners('sitemap-change');
        this.updateStyles();
        this.getSitemap();
    },

    reportTabVisibility : function(bVisible) {
        this.bVisible = bVisible;
        // TODO: Drop this feature altogether?
    },

    informListeners : function(sReason) {
        if (true)
            this.lib.debug('inform ' + this.aListeners.length +
                           ' Listeners(' + sReason + ')');
        var bReload = this.bReloadNext;
        function makeFunc(el) {
            return function() {
                el(bReload, sReason);
            };
        }
        for each (var el in this.aListeners) {
            setTimeout(makeFunc(el), 0);
        }
    },

    // Report that a client is making the sitemap available to the
    // user.  iLevel==0 means that there is no sitemap.  sKey
    // identifies the client (e.g. sidebar, toolbar-button).  Whenever
    // a value is changed, this function goes through all recorded
    // values for the tab and finds the maximum.
    setDisplayLevel : function(sKey, iLevel) {
        // Record this level.
        this.mDisplayLevel[sKey] = iLevel;

        this.updateStyles();
    },

    updateStyles : function(oSitemap) {
        // Get the highest level, and detect any changes from the
        // previous maximum.
        var max = 0;
        for each (var i in this.mDisplayLevel) {
            if (i > max)
                max = i;
        }
//         this.lib.debug('New level: ' + max +
//                     '\nOld level: ' + this.iLastDisplayLevel);
        if (false && !oSitemap && max == this.iLastDisplayLevel)
            return;
        // There was a change.  We might have to enable or disable
        // some styles.

        if (this.bNewPage) {
            // The page corresponding to the current URI hasn't loaded
            // yet, so we can't update its contents.
            this.lib.debug('Too soon to change classes: page not loaded');
            return;
        }

        if (false) {
            // Use the URIs specified in the stylesheet.  (This is the
            // old way.)

            var styleChange;
            var self = this;
            if (max >= 50) {
                styleChange = function(oSitemap) {
                    if (!oSitemap) {
                        self.applyStyles(null);
                        return;
                    }

                    // Apply the sitemap-specified styles.
                    self.applyStyles(oSitemap.sStyleSheetData);
                };
            } else {
                styleChange = function(oSitemap) {
                    // Withdraw the sitemap-specified styles.
                    self.applyStyles(null);
                };
            }
            // Make sure any changes are applied as soon as the
            // sitemap is loaded.
            if (oSitemap)
                styleChange(oSitemap);
            else
                this.getSitemap(styleChange);
        } else if (false) {
            // TODO: Report the new level to the Media Query engine.

            // It should be as easy as this (fantasy code)!!:

            this.mqEngine.setFeatureNS(browser,
                                       'http://standard-sitemap.org/',
                                       'level',
                                       max);
        } else {
            // Add/remove a family of classes on an element of the
            // displayed document.  For example, if the class root is
            // womble, and the level is set to 20, ensure that the
            // classes womble-over-0 to womble-over-19 are present,
            // and that womble-over-20 to womble-over-100 are absent.
            // Similarly, add womble-under-21 to womble-under-100, and
            // remove womble-under-0 to womble-under-20.
            this.updateStyleClasses(max);
        }

        // Record the new display level.
        this.iLastDisplayLevel = max;
    },

    updateStyleClasses : function(level) {
        var doc = this.oBrowser.contentDocument;
        if (!doc) {
            if (false)
                this.lib.debug("Level not set to " + level
                               + "; no contentDocument");
            return;
        }
        if (!this.oSitemap) {
            if (false)
                this.lib.debug("Level not set to " + level
                               + "; no sitemap");
            return;
        }
        this.lib.debug("Sitemap " + this.oSitemapURI.spec +
                       " classes applied to " +
                       this.oBrowser.currentURI.spec +
                       "; sitemap URI is " + this.oSitemap.oSitemapURI.spec);
        if (this.oSitemap.oSitemapURI.spec != this.oSitemapURI.spec)
            this.lib.debug("*** URI mismatch ***");
        this.oSitemap.ApplyClassChanges(0, 100, level, doc, window);
    },

    getVisibility : function() {
        return this.bVisible;
    },

    sLastAddedStyle : null,

    applyStyles : function(sAddedStyle) {
        if (sAddedStyle == this.sLastAddedStyle)
            return;

        if (true) {
            if (this.sLastAddedStyle)
                this.lib.debug('Dropping style ' + this.sLastAddedStyle);
            if (sAddedStyle)
                this.lib.debug('Adding style ' + sAddedStyle);
        }

        var oSSS =
            Components.classes["@mozilla.org/content/style-sheet-service;1"]
            .getService(Components.interfaces.nsIStyleSheetService);
        var oIOService =
            Components.classes["@mozilla.org/network/io-service;1"]
            .getService(Components.interfaces.nsIIOService);

        if (this.sLastAddedStyle) {
            var oURI = oIOService.newURI(this.sLastAddedStyle, null, null);
            if (oSSS.sheetRegistered(oURI, oSSS.USER_SHEET))
                oSSS.unregisterSheet(oURI, oSSS.USER_SHEET);
        }

        if (sAddedStyle) {
            var oURI = oIOService.newURI(sAddedStyle, null, null);
            if (!oSSS.sheetRegistered(oURI, oSSS.USER_SHEET))
                oSSS.loadAndRegisterSheet(oURI, oSSS.USER_SHEET);
        }

        this.sLastAddedStyle = sAddedStyle;
    },

};

// Start things up when a window is created.
addEventListener('load',
                  function(event) {
                     window["org.standard-sitemap.monitor"].start();
                 }, false);
addEventListener('unload',
                  function(event) {
                     window["org.standard-sitemap.monitor"].end();
                 }, false);
