// -*- c-basic-offset: 4; indent-tabs-mode: nil -*-

/* ***** BEGIN LICENSE BLOCK *****
Version: MPL 1.1/GPL 2.0/LGPL 2.1

The contents of this file are subject to the Mozilla Public License
Version 1.1 (the "License"); you may not use this file except in
compliance with the License. You may obtain a copy of the License at
http://www.mozilla.org/MPL/

Software distributed under the License is distributed on an "AS IS"
basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
License for the specific language governing rights and limitations
under the License.

The Original Code is Standard-Sitemap Navigator code.

The Initial Developer of the Original Code is the
Standard-Sitemap Team.

Contributor(s): (see license.txt)

Alternatively, the contents of this file may be used under the terms of
either the GNU General Public License Version 2 or later (the "GPL"), or
the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
in which case the provisions of the GPL or the LGPL are applicable instead
of those above. If you wish to allow use of your version of this file only
under the terms of either the GPL or the LGPL, and not to allow others to
use your version of this file under the terms of the MPL, indicate your
decision by deleting the provisions above and replace them with the notice
and other provisions required by the GPL or the LGPL. If you do not delete
the provisions above, a recipient may use your version of this file under
the terms of any one of the MPL, the GPL or the LGPL.
***** END LICENSE BLOCK ***** */

/*ToDo's moved from old sitemap.js
  TODO: Validate doc against schema so we don't have to do endless validations here.
  TODO: cache and re-use sitemaps.
*/



var EXPORTED_SYMBOLS = [];


Components.utils.import("resource://modules.standard-sitemap.org/sslib.jsm");


var Application = Components.classes["@mozilla.org/fuel/application;1"]
    .getService(Components.interfaces.fuelIApplication);



/*StandardSitemapService is a service object for processing standard sitemaps*/
function StandardSitemapService()
{
    this.Initialize();
}

StandardSitemapService.prototype =
{
    lib : Application.storage.get("org.standard-sitemap.library", null),

    Sitemap : function(window, sSitemapURI,
                       fCallbackFunction, bSkipCache) {
        this.Initialize(window, sSitemapURI,
                        fCallbackFunction, bSkipCache);
    },

    aSitemapCache: null, //Array of previously loaded sitemaps.
    
    /*Initialize the service.*/
    Initialize: function()
    {
        this.aSitemapCache = {};
    },
    
    /*Reset all properties*/
    Reset: function()
    {
        for each(oSitemap in this.aSitemapCache) oSitemap.Reset();
    },

    LoadSitemapInternal: function(window,
                                  sSitemapURI, fCallbackFunction, bReload,
                                  bSkipCache)
    {
        var self = this;
        if (!bReload) {
            var oSitemap = this.aSitemapCache[sSitemapURI];
            if (oSitemap) {
                window.setTimeout(function() {
                    if (false)
                        self.lib.debug("Retrieved sitemap for " + sSitemapURI);
                    fCallbackFunction(oSitemap);
                }, 0);
                return;
            }
        }
        
        var oSitemapService = this;
        var fServiceCallback = function(oLoadedSitemap)
        {
            if (oLoadedSitemap && fCallbackFunction)
            {
                if (false)
                    self.lib.debug("Obtained sitemap for " + sSitemapURI +
                                   (oLoadedSitemap ? "" : " (null)"));
                oSitemapService.aSitemapCache[sSitemapURI] = oLoadedSitemap;
                fCallbackFunction(oLoadedSitemap);
            }
        }
        new (this.Sitemap)(window, sSitemapURI,
                                   fServiceCallback, bSkipCache);
    },
    
    /*Checks if a sitemap object for the given url already exists, if not it will initialize a new one and cache it
      The callback function will be executed when the sitemap is done loading or immediately if we found a cached version.
    */
    LoadSitemap: function(window,
                          sSitemapURI, fCallbackFunction)
    {
        this.LoadSitemapInternal(window,
                                 sSitemapURI, fCallbackFunction,
                                 false, false);
    },
    
    /*Reload a sitemap and optionally bypass bypass the cache*/
    ReLoadSitemap: function(window,
                            sSitemapURI, fCallbackFunction, bSkipCache)
    {
        this.LoadSitemapInternal(window,
                                 sSitemapURI, fCallbackFunction,
                                 true, bSkipCache);
    },
    
    /*Gets a sitemap directly from the cache.*/
    GetSitemapFromCache: function(sSitemapURI)
    {
        return this.aSitemapCache[sSitemapURI];
    },
    
    /*Check if a given sitemap exists in the cache*/
    SitemapExists: function(sSitemapURI)
    {
        return !!this.aSitemapCache[sSitemapURI];
    }
}

// We give each sitemap object a unique id.  A monotonically
// increasing integer should do fine, as it's not likely to wrap
// around for ages.
var nextId = 0;

/*Sitemap is an object to provide easy processing of standard sitemap files*/
StandardSitemapService.prototype.Sitemap.prototype = {
    lib : Application.storage.get("org.standard-sitemap.library", null),

    Logo : function(oImageVariants, oLinkVariants) {
        this.SetProperties(oImageVariants, oLinkVariants);
    },

    Node : function(creator) {
        this.creator = creator;
        this.Initialize();
    },

    VariantData : function(oElement, mDefaults) {
        this.mEntries = { };
        this.mAddrs = { };
        this.aLangs = { };
        this.decodeElement({ }, oElement, mDefaults);
    },

    aClassChanges: [],
    fCallbackFunction: null, //Callback function to call when loading is completed.
    oSitemapNode: null, //Root node containing the Nodes
    oSitemapURI: null, //nsIURI of the processed sitemap
    oStyleSheetURI: null, //The stylesheet we need to apply to the website
    oReplacesStyleSheetURI: null, //The stylesheet being replaced
    oSitemapLogo: null, //Processed Logo item
    oHome: null, // Holds details of the Home page.
    oContact: null, // Holds details of the Contact page.
    oContentInfo: null, // Holds details of the ContentInfo page.
    oLogoItem: null, // Content of the item providing the logo link
    mParsedElements: null, // Mapping elements to sitemap nodes
    mURIsToNodes: null, // For finding nodes by URI

    //TODO: This needs to be changed if we want to support multiple search pages.
    oSearch: null, // Holds details of site search.
    oSearchPage: null, // Holds details of the Search page.
    oSitemapBase: null, // base URI of document element
    bIsGoogleSiteSearch: false, //If the sitemap does not describe a role=search, google is used.
    bNoAutoFind : false, // True if visiting a page should not cause the item to be sought in the tree.
    
    aRequestCache: null, //Array of previously loaded sitemap xml files to save multiple requests to load the same sitemap.

    sLanguage: "", //The sitemap language.
    aAvailableLanguages: {}, // map(string->boolean) of languages detected
    sType: "text/html", //The sitemap content type.
    sCharset: "UTF-8", //The sitemap charset.
    sStyleSheetData: "", //Data url containing the css.
    sReplacesStyleSheetData: "", //Data url containing the @replace css.
    sSidebarWidth: "18em", //Store the width of the sidebar per session...
    
    bSkipCache: false, //Should we bypass the cache when making requests.
    nextHashCode : 0,

    /*Initialize the object using the given uri, when finished loading the callback function will be called.*/
    Initialize: function(window, sSitemapURI,
                         fCallbackFunction, bSkipCache)
    {
        //ToDo: Should we do anything when initialized without a URI?
        if (!sSitemapURI) return;

        this.id = nextId++;

        //ToDo: If we decide not to keep all xml's cached, then the if statement can go (always init a new object)
        if (!this.aRequestCache) this.aRequestCache = {};
        this.bSkipCache = bSkipCache;

        var oSitemap = this;
        var hashFunc = function(obj) {
            if ("standardSitemapHashCode" in obj)
                return obj.standardSitemapHashCode;
            return obj.standardSitemapHashCode = oSitemap.nextHashCode++;
        };
        this.mParsedElements = this.lib.newMap(4001, hashFunc);
        
        this.fCallbackFunction = fCallbackFunction;
        this.oSitemapURI = this.lib.makeURI(sSitemapURI);
        this.oSitemapLogo = new (this.Logo)();
        this.oLogoItem = null;
        this.mURIsToNodes = { };
        this.aClassChanges = [];
        this.aAvailableLanguages = {};

        //Setup a callback function to call when the request is finished.
        var fCallback = function(oEntry)
        {
            oSitemap.ProcessSitemapDocument(window,
                                            oEntry.oRequest.responseXML,
                                            oEntry.oSitemapURI);
        }
        //Make the request for the xml document.
        this.RequestSitemapDocument(window, this.oSitemapURI, fCallback);
        
    },
    
    /*Reset all properties*/
    Reset: function()
    {
        this.aClassChanges = [];
        this.aAvailableLanguages = {};
        this.fCallbackFunction = null;
        this.oSitemapURI = null;
        this.oStyleSheetURI = null;
        this.oReplacesStyleSheetURI = null;
        if (this.oSitemapLogo) this.oSitemapLogo.Reset();
        this.oSitemapBase = null;
        this.oSitemapLogo = null;
        this.oHome = null;
        this.oContact = null;
        this.oContentInfo = null;
        this.oSearch = null;
        this.oLogoItem = null;
        this.bIsGoogleSiteSearch = false;
        this.bNoAutoFind = false;
        this.mParsedElements = null;
        if (this.oSitemapNode) this.oSitemapNode.Reset();
        this.oSitemapNode = null;
        this.mURIsToNodes = null;
        //ToDo: Should this also reset the cache of all sitemaps or just the current one?
        //this.aRequestCache = null;

        this.sLanguage = "";
        this.sType = "";
        this.sCharset = "";
        this.sStyleSheetData = "";
        this.sReplacesStyleSheetData = "";
        
        this.bSkipCache = false;
    },

    /*Processes the sitemap xml document*/
    ProcessSitemapDocument: function(window, oXMLDocument, oSitemapURI)
    {
        this.mParsedElements.clear();

        var oSitemapDocument =
            this.lib.getElementsByTagNameSM(oXMLDocument, 'sitemap')[0];

        //ToDo: Add useful information to the root item? It's a basically just a placeholder now.
        this.oSitemapNode = new (this.Node)(this.Node);
        this.oSitemapNode.oSitemapURI = oSitemapURI;
        this.oSitemapNode.sStateId = oSitemapDocument ?
            oSitemapDocument.ownerDocument.documentURI : oSitemapURI.spec;
        var oBase = oSitemapDocument ?
            this.lib.getBaseURI(oSitemapDocument) : oSitemapURI;
        this.oSitemapNode.
        SetProperties("sitemap",
                      "Root item for the sitemap", "",
                      oBase.spec,
                      "", "1", "", false);
        
        this.ProcessStyleItems(window, oSitemapDocument);
        this.ProcessClassChanges(oSitemapDocument);

        // Detect site-wide options.
        if (oSitemapDocument) {
            var sOptions = oSitemapDocument.getAttribute('opts');
            if (sOptions) {
                var aOptions = sOptions.toString().split(/\s+/);
                for each (var sOption in aOptions) {
                    //this.lib.dump('Option detected: ' + sOption);
                    switch (sOption) {
                    case 'noautofind':
                        this.bNoAutoFind = true;
                        break;
                    default:
                        this.lib.dump('Option ignored: ' + sOption);
                        break;
                    }
                }
            }
        }

        // Optionally, keep track of the number of outstanding
        // sitemaps being handled.  Only when this returns to zero are
        // we ready to invoke the callback.
        var oSitemap = this;
        var iOutstanding = 0;
        var fIncrement = null, fDecrement = null;
        if (this.fCallbackFunction) {
            var fCBF = this.fCallbackFunction;
            fIncrement = function() {
                iOutstanding++;
                //oSitemap.lib.dump('Loading another doc... ' + iOutstanding);
            };
            fDecrement = function() {
                //oSitemap.lib.dump('Finished a doc... ' + iOutstanding);
                if (--iOutstanding == 0) {
                    // Call the function so the caller can start
                    // accessing the sitemap.

                    //oSitemap.lib.dump('All finished');
                    fCBF(oSitemap);
                }
            };
        }

        // We're processing one sitemap right now.
        if (fIncrement) fIncrement();

        this.ProcessSitemapNodes(window, this.oSitemapNode,
                                 oSitemapDocument, fIncrement, fDecrement);

        /*Implement Google site search if no role=search was found*/
        if (!this.oSearch || !this.oSearch.getAddress())
        {
            var mSearch = {
                "name" : this.lib
                    .TranslateString("sitemap.search.google.label"),
                "desc" : this.lib
                    .TranslateString("sitemap.search.google.description"),
                "searchdata" : this.lib.getCharPref('defaultSearchQuery'),
                "addr" : this.lib
                    .makeURI(this.lib.getCharPref('defaultSearchSite'))
            };

            this.oSearch = new (this.VariantData)(null, mSearch);
            this.bIsGoogleSiteSearch = true;
        }
        this.oSitemapBase = oBase;

        // Process the logo after having checked for @role=logo.
        //this.ProcessLogo(oSitemapDocument);

        // We've finished processing this sitemap, but other
        // referenced sitemaps might still be in process.
        if (fDecrement) fDecrement();
    },

    ReplaceFields : function(sFormat, sQuery) {
        var sResult = '';
        var iPos = 0;

        while (iPos < sFormat.length) {
            //this.lib.dump("Remaining is [" + sFormat.substring(iPos) + "]");

            // Find the next "%".
            var iPer = sFormat.indexOf("%", iPos);
            if (iPer < 0) {
                sResult += sFormat.substring(iPos);
                break;
            }

            // Skip ordinary characters up to that "%".
            sResult += sFormat.substring(iPos, iPer);
            iPos = iPer;

            var text;
            switch (sFormat.charAt(iPos + 1)) {
            case '0': case '1': case '2': case '3': case '4':
            case '5': case '6': case '7': case '8': case '9':
            case 'a': case 'b': case 'c': case 'd': case 'e': case 'f':
            case 'A': case 'B': case 'C': case 'D': case 'E': case 'F':
                // Skip %-encoded characters.
                sResult += sFormat.substring(iPos, iPos + 3);
                iPos += 3;
                continue;

            case '{':
                // Process embraced field.
                iPer = sFormat.indexOf("}", iPos + 2);
                if (iPer < 0) {
                    // It wasn't closed.
                    sResult += sFormat.substring(iPos);
                    iPos = sFormat.length;
                    break;
                }
                text = sFormat.substring(iPos + 2, iPer);
                iPos = iPer + 1;
                break;

            default:
                // Process single-char field.
                text = sFormat.charAt(iPos + 1);
                iPos += 2;
                break;
            }

            //this.lib.dump("Field is [" + text + "]");

            if (text == "s") {
                sResult += escape(sQuery);
            } else if (text == "r") {
                sResult = sResult +
                    escape(this.oSitemapBase.resolve('/'));
            } else if (text == "h") {
                sResult = sResult +
                    escape(this.oSitemapBase.host);
            } else if (text.lastIndexOf("w") == text.length - 1) {
                var iUp = parseInt(text.substring(0, text.length - 1));
                var sUp = "./";
                while (iUp-- > 0)
                    sUp += "../";
                sResult = sResult +
                    escape(this.oSitemapBase.resolve(sUp));
            } else {
                sResult += "%{" + text + "}";
            }
        }

        return sResult;
    },

    /*Proces all childnodes of oParentNode and store them in the sitemap object*/
    ProcessSitemapNodes: function(window, oSitemapNode, oParentNode,
                                  fIncrement, fDecrement)
    {
        if (!oParentNode) return;
        this.ProcessRemainingNodes(window, oSitemapNode,
                                   oParentNode.childNodes,
                                   fIncrement, fDecrement, 0);
    },

    // TODO: Instead of always using oSitemapNode as the parent, allow
    // another node to be passed in as the 'effective parent', in
    // cases where oSitemapNode doesn't have a valid URI to visit or a
    // valid name to display.
    ProcessRemainingNodes : function(window, oSitemapNode, aChildNodes,
                                     fIncrement, fDecrement,
                                     iIndex, oPrevNode) {
        // Process <item> and <group> elements immediately, but set up
        // a call-back to handle <external>s.  In either case, make
        // sure that the new node and its previous sibling link to
        // each other.
        for ( ; iIndex < aChildNodes.length; iIndex++) {
            var oChildNode = aChildNodes[iIndex];
            if (oChildNode.nodeType != oChildNode.ELEMENT_NODE) continue;
            switch (oChildNode.tagName) {
            default:
                break;

            case 'item':
            case 'group':
                var oNewNode =
                    this.ProcessSitemapNode(window,
                                            oSitemapNode, oChildNode,
                                            fIncrement, fDecrement);
                if (oNewNode) {
                    if (oPrevNode) {
//                         this.lib.dump('Linking ' + oPrevNode.GetName() +
//                                ' to ' + oNewNode.GetName() +
//                                ' under ' + oSitemapNode.GetName());
                        oPrevNode.mNext.put(oSitemapNode, oNewNode);
                        oNewNode.mPrev.put(oSitemapNode, oPrevNode);
                    }
                    oPrevNode = oNewNode;
                }
                break;

            case 'external':
                var oldThis = this;
                var fNext = function(oNewNode) {
                    if (oNewNode) {
                        if (oPrevNode) {
                            oPrevNode.mNext.put(oSitemapNode, oNewNode);
                            oNewNode.mPrev.put(oSitemapNode, oPrevNode);
                        }
                        oPrevNode = oNewNode;
                    }
                    oldThis.ProcessRemainingNodes(window,
                                                  oSitemapNode, aChildNodes,
                                                  fIncrement, fDecrement,
                                                  iIndex + 1,
                                                  oNewNode ?
                                                  oNewNode : oPrevNode);
                };
                this.ProcessSitemapExternalNode(window,
                                                oSitemapNode, oChildNode,
                                                fIncrement, fDecrement,
                                                fNext);
                return;
            }
        }
    },

    GetLanguages : function() {
        var r = [];
        for (var lang in this.aAvailableLanguages)
            r.push(lang);
        return r;
    },

    // Proces a single <group> or <item> node 'oElement', and add it
    // under the sitemap item 'oParentSitemapNode'.
    ProcessSitemapNode: function(window, oParentSitemapNode, oElement,
                                 fIncrement, fDecrement)
    {
        // Have we already processed this element?
        var oExisting = this.mParsedElements.getNull(oElement);
        if (oExisting) {
            // Yes, but what was the result?
            oExisting = oExisting.value;

            if (oExisting) {
                // An actual node was created before, so make sure
                // that it knows it has another parent.
                oExisting.mParent.put(oParentSitemapNode, true);
                oParentSitemapNode.aChildNodes.push(oExisting);
            } else {
                // Nothing was generated as a result.
            }
            return oExisting;
        }

        // Determine an identifier for the node so we can persist its
        // state.
        var sourceId = oElement.getAttributeNS(this.lib.xmlNS, "id");
        if (sourceId && sourceId != "") {
            sourceId =
                this.lib.makeURI(oElement.ownerDocument.documentURI)
                .resolve('#' + sourceId);
        } else {
            sourceId = null;
        }

        // Record that we have processed this element, and indicate
        // failure to create a sitemap node by default.  We override
        // this later.
        this.mParsedElements.put(oElement, null);

        // Assimilate the negotiable content of this item.
        var oVariants = new (this.VariantData)(oElement);
        for (var lang in oVariants.getLangs())
            this.aAvailableLanguages[lang] = true;

        /* Check the roles first.  They might imply whether to include
           the item in the navigation tree. */
        var bAllRolesExtracted = true;

        // We assume a default role of 'tree' - the item should appear
        // in the tree, along with any items implied by this element's
        // children.
        var sRole = oElement.getAttribute('role');
        if (!sRole)
            sRole = 'tree';

        // Now go through each role.  TODO: Use proper regexp here?
        var aRoles = this.lib.stringTrim(sRole).split(/\s+/)
        // TODO: Just use sRole.trim() if we drop support for FF3.0.10
        // (or wherever trim() was introduced).
        for each (var sRole in aRoles) {
            switch (sRole.toLowerCase()) {
            case 'tree':
            default:
                // We don't recognise this role, so we can't risk
                // removing it from the tree.
                bAllRolesExtracted = false;
                break;

            case 'contact':
                // Don't recognise this role unless an address is
                // given.
                if (!oVariants.getAddress())
                    bAllRolesExtracted = false;
                else
                    this.oContact = oVariants;
                break;

            case 'home':
                // Don't recognise this role unless an address is
                // given.
                if (!oVariants.getAddress())
                    bAllRolesExtracted = false;
                else
                    this.oHome = oVariants;
                break;

            case 'contentinfo':
                // Don't recognise this role unless an address is
                // given.
                if (!oVariants.getAddress())
                    bAllRolesExtracted = false;
                else
                    this.oContentInfo = oVariants;
                break;

            case 'search':
                // Don't recognise this role unless an address is
                // given.
                if (!oVariants.getAddress())
                    bAllRolesExtracted = false;
                else
                    this.oSearch = oVariants;
                break;

            case 'searchpage':
                // Don't recognise this role unless an address is
                // given.
                if (!oVariants.getAddress())
                    bAllRolesExtracted = false;
                else
                    this.oSearchPage = oVariants;
                break;

            case 'logo':
                // Don't recognise this role unless an address is
                // given.
                if (!oVariants.getAddress())
                    bAllRolesExtracted = false;
                else
                    this.oLogoItem = oVariants;

                // We don't really recognise this role any more, as
                // the logo has been disabled.
                bAllRolesExtracted = false;
                break;
            }
        }

        // We can't include an unnamed non-group item in the tree.
        if (oElement.tagName == 'item' && !oVariants.getName())
            return;

        // Determine whether we include this item in the tree.
        switch (oElement.getAttribute('tree')) {
        case 'include':
            break;

        case 'exclude':
            return;

        case 'user':
        default:
            // The user may prefer to keep items deemed optional so
            // far.
            if (this.lib.getBoolPref('showRoleItemsInTree'))
                break;
        case 'auto':
            // We also can't drop optional items if they have any
            // roles we didn't understand.
            if (bAllRolesExtracted)
                return;
            break;
        }

        // Create an internal node associated with this XML element,
        // but check for recursion by inspecting ancestors for the
        // same XML element.
        var oSitemapNode = oParentSitemapNode.AddChildNode(oElement);
        if (!oSitemapNode)
            return;
        // Give the node an id.
        oSitemapNode.sStateId = sourceId;

        this.mParsedElements.put(oElement, oSitemapNode);

        // Get all the URIs and record them so we can find this node
        // quickly.
        for (var addr in oVariants.mAddrs) {
            var nodeList = this.mURIsToNodes[addr];
            if (!nodeList) {
                nodeList = [];
                this.mURIsToNodes[addr] = nodeList;
            }
            nodeList.push(oSitemapNode);
        }

        var sRelation = oElement.getAttribute('relation');
        var sPriority = oElement.getAttribute('priority');

        // Treat the first <item> as the logo address.
        if (!this.oLogoItem)
            this.oLogoItem = oVariants;

        // Treat the first <item> as the home address.
        if (!this.oHome)
            this.oHome = oVariants;

        //Add standard information
        oSitemapNode.SetProperties(oVariants, sRole, sPriority, sRelation,
                                   (oElement.tagName == "group"));

        // Get the @order attribute, which tells us how we can
        // automatically group children of this element if it creates
        // GUI structures which are too tall.
        oSitemapNode.autoGroupOrder = oElement.getAttribute('order');
        if (oSitemapNode.autoGroupOrder == '' || !oSitemapNode.autoGroupOrder)
            oSitemapNode.autoGroupOrder = 'none';

        //Process this item's children again.
        this.ProcessSitemapNodes(window, oSitemapNode, oElement,
                                 fIncrement, fDecrement);

        return oSitemapNode;
    },

    /* Should we indeed also check for the URI upto the # in here?
    SS: No, we're getting the list of exactly matching nodes. */
    GetNodesForURI : function(sURI) {
        var r = this.mURIsToNodes[sURI];
        //if (!r) r = this.mURIsToNodes[sURI.split("#")[0]];
        return r ? r : [];
    },

    // Proces an external XML node 'oElement', and add it as a child
    // of 'oParentSitemapNode'.  Call 'fIncrement' once before
    // starting anything asynchronous, and 'fDecrement' similarly when
    // such a task is complete - these are used to tell the caller
    // when all work is done.  'fDone' must be called with the new
    // sitemap node that was created.
    ProcessSitemapExternalNode: function(window,
                                         oParentSitemapNode, oElement,
                                         fIncrement, fDecrement, fDone)
    {
        //Split the URI-Reference href into a plain URI (the location of the external document) and the anchor (the idref).
        var oURI = this.lib.getURIAttrNS(oElement, this.lib.xlinkNS, 'href');
        if (!oURI)
            oURI = this.lib.getURIAttr(oElement, 'url');
        var oSitemapURI = this.lib.makeURI(oURI.resolve(""));

        // Get the fragment identifier.
        var sFragment = "";
        var iHashPos = oURI.spec.indexOf('#');
        if (iHashPos > -1)
        {
            // The fragment identifier appears after the first hash.
            sFragment = oURI.spec.substring(iHashPos + 1);

            // Decode any %xx characters in the fragment identifier.
            sFragment = unescape(sFragment);
        }

        var oSitemap = this;
        var fCallback = function(oEntry)
        {
            // Do not return from this function without calling
            // fDecrement() (if set).

            // Same goes for fDone(), but it is allowed to pass it on
            // to another call.

            var oXMLDocument = oEntry.oRequest.responseXML;
            var oSitemapDocument =
                oSitemap.lib.getElementsByTagNameSM(oXMLDocument, 'sitemap')[0];
            if (sFragment)
            {
                var oElement = oEntry.GetID(sFragment);
                if (oElement) {
                    // Process this element immediately if it's an
                    // <item> or a <group>, defer to another call to
                    // this method if it's <external>, else ignore.
                    switch (oElement.tagName) {
                    default:
                        if (fDone) fDone();
                        break;

                    case 'external':
                        oSitemap.ProcessSitemapExternalNode(window,
                                                            oParentSitemapNode,
                                                            oElement,
                                                            fIncrement,
                                                            fDecrement,
                                                            fDone);
                        break;

                    case 'item':
                    case 'group':
                        var oNewNode =
                            oSitemap.ProcessSitemapNode(window,
                                                        oParentSitemapNode,
                                                        oElement,
                                                        fIncrement, fDecrement);
                        if (fDone) fDone(oNewNode);
                        break;
                    }
                } else {
                    if (fDone) fDone();
                    oSitemap.lib.dump('item ' + sFragment + ' not found in '
                                      + oXMLDocument.documentURI);
                }
            } else {
                // TODO: This is a bit dodgy.  If the author
                // specifies an entire sitemap, do we add one
                // item to represent the whole sitemap?  Or do
                // we (as below) add all its children.  I
                // favour the former, but I haven't worked out
                // how to do it yet.  Anyway, this might be a
                // rare occurance.
                //childCount +=
                //Sitemap.addChildren(parent, docBase, doc.childNodes[0]);

                // Make sure we indicate that it's safe to add more
                // children to the parent.
                if (fDone) fDone();
            }
            if (fDecrement) fDecrement();
        };
        if (fIncrement) fIncrement();
        this.RequestSitemapDocument(window, oSitemapURI,
                                    fCallback, fDecrement);
    },

    ApplyClassChanges : function(iMin, iMax, iVal, doc, window) {
        //this.lib.debug('CLASS CHANGES');
        for each (var cc in this.aClassChanges) {
            //this.lib.debug('  Setting ' + cc.prefix + ' to ' + iVal);
            this.lib.setClassScale(cc.elemPath, cc.attrName, cc.attrNS,
                                   cc.resolver, cc.prefix,
                                   iMin, iMax, iVal, doc, window);
        }
    },

    ProcessClassChanges : function(oSitemapDocument) {
        if (!oSitemapDocument) return;

        // Get <class-change elem="xpath" attr="class attr name"
        // prefix="class-name-prefix" />.
        //this.lib.debug('Extracting class-changes');
        var elems =
            this.lib.getElementsByTagNameSM(oSitemapDocument, "class-change");
        for (var i = 0; i < elems.length; i++) {
            var elem = elems.item(i);
            var cc = { };
            cc.elemPath = elem.getAttribute("elem");
            cc.attrName = elem.getAttribute("attr");
            if (cc.attrName == "") {
                cc.attrName = "class";
                cc.attrNS = null;
            } else {
                var idx = cc.attrName.indexOf(':');
                if (idx >= 0) {
                    var attrns = cc.attrName.substring(0, idx);
                    cc.attrName = cc.attrName.substring(idx + 1);
                    cc.attrNS = elem.lookupNamespaceURI(attrns)
                } else {
                    cc.attrNS = null;
                }
            }
            cc.prefix = elem.getAttribute("prefix");
            //this.lib.debug('Sitemap defines class ' + cc.prefix);
            cc.resolver = function (s) {
                return elem.lookupNamespaceURI(s);
            };
            this.aClassChanges.push(cc);
        }
    },

    /*Gets the style related items from the document*/
    ProcessStyleItems: function(window, oSitemapDocument)
    {
        if (!oSitemapDocument) return;
        var oElement, oBaseURI;
        var iIndex, iLen = oSitemapDocument.childNodes.length;
        
        for (iIndex = 0; iIndex < iLen; iIndex ++)
        {
            oElement = oSitemapDocument.childNodes[iIndex];
            if (oElement.nodeType != oElement.ELEMENT_NODE)
                continue;
            oBaseURI = this.lib.getBaseURI(oElement);
            if (oElement.tagName == "style") {
                this.oStyleSheetURI = this.lib.getURIAttrNS(oElement, this.lib.xlinkNS, 'href');
                if (!this.oStyleSheetURI) this.oStyleSheetURI = this.lib.getURIAttr(oElement, 'url');
                
                this.sStyleSheetData = this.lib.ReadCSS(window, this.oStyleSheetURI.spec);

                var aPath = this.oSitemapURI.spec.split("/");
                var sDocument = this.oSitemapURI.spec.replace(aPath[aPath.length - 1], "");

                if (this.sStyleSheetData && this.sStyleSheetData != "")
                {
                    this.sStyleSheetData = "data:text/css,@namespace url('http://www.w3.org/1999/xhtml');\n@-moz-document url-prefix('" + sDocument + "') {\n" + this.sStyleSheetData + "\n}";
                }

                if (oElement.getAttribute('replace')) {
                    this.oReplacesStyleSheetURI = this.lib.getURIAttr(oElement, "replace");
                    // Since resolveURI will resolve an empty string
                    // to the base uri we would later on have a
                    // problem trying to re-register that URI as a
                    // stylesheet...

                    // SS: Surely not a problem: if @replace is not
                    // set, this block is not executed.  replace=""
                    // and no @replace don't have to be taken as
                    // equivalent.
                    if (this.oReplacesStyleSheetURI.equals(oBaseURI)) this.oReplacesStyleSheetURI = null;
                    if (this.oReplacesStyleSheetURI)
                    {
                        this.sReplacesStyleSheetData = this.lib.ReadCSS(window, this.oReplacesStyleSheetURI.spec);
                        if (this.sReplacesStyleSheetData && this.sReplacesStyleSheetData != "")
                        {
                            this.sReplacesStyleSheetData = "data:text/css,@namespace url('http://www.w3.org/1999/xhtml');\n@-moz-document url-prefix('" + sDocument + "') {\n" + this.sReplacesStyleSheetData + "\n}";
                        }
                    }
                }
            }
        }
    },

    /*Gets the logo from the document*/
    ProcessLogo: function(oSitemapDocument)
    {
        if (!oSitemapDocument) return;

        var oElement;
        var iIndex, iLen = oSitemapDocument.childNodes.length;
        
        for (iIndex = 0; iIndex < iLen; iIndex ++)
        {
            oElement = oSitemapDocument.childNodes[iIndex];
            if (oElement.nodeType != oElement.ELEMENT_NODE)
                continue;
            if (oElement.tagName == "image") {
                var oImageVariants = new (this.VariantData)(oElement);
                if (!oImageVariants.getAddress())
                    continue;
                this.oSitemapLogo =
                    new (this.Logo)(oImageVariants, this.oLogoItem);
            }
        }
    },
    
    /*Clears the cached entry for the given URI, should be called when the page is refreshed?*/
    ClearCache: function(sSitemapURI)
    {
        var oURI = this.lib.makeURI(sSitemapURI);
        this.aRequestCache[oURI.spec] = null;
    },

    /* Sends an HTTP request for the sitemap and executes the callback
     * function if successful.  fError is called if and only if
     * fCallback is not. */
    RequestSitemapDocument: function(window, oSitemapURI,
                                     fCallback, fError)
    {
        //Check if the document has already been loaded in this particular construction of the sitemap.
        var oEntry = this.aRequestCache[oSitemapURI.spec];

        // Define what to do when the request is complete.
        var oSitemap = this;
        var fOnCompletion = function() {
            if (oEntry.bOkay) {
                fCallback(oEntry);
            } else if (fError) {
                fError();
            }
        };

        if (oEntry)
        {
            //The request has already been made.
            if (oEntry.oRequest.readyState == 4) {
                // The request has already finished, so call the event
                // handler almost immediately.  We could call it
                // directly, but that causes the call stack to
                // increase drastically when there are a lot of
                // <external>s.  The call below allows this function
                // to return, leaving fOnCompletion as a task for
                // later.
                window.setTimeout(fOnCompletion, 0);
            } else {
                // The request has not yet finished, so push another
                // action to be performed.
                oEntry.aActions.push(fOnCompletion);
            }
            return;
        }

        // The request has not yet been made, so prepare one.
        //this.lib.dump('Making request for ' + oSitemapURI.spec);
        var oRequest = new window.XMLHttpRequest();
        oRequest.overrideMimeType('text/xml');

        // Make an entry for this request.
        var self = this;
        oEntry =
        {
            GetID: function(sID) {
                // Try to get xml:id first.
                var res = this.oRequest.responseXML
                    .evaluate('//*[@xml:id="'+sID+'"]',
                              this.oRequest.responseXML,
                              function () { return self.lib.xmlNS; },
                              window.XPathResult.FIRST_ORDERED_NODE_TYPE,
                              null).singleNodeValue;
                if (res)
                    return res;

                // Otherwise, get plain old id.

                // TODO: Disable this later?
                for each (var cand in self.lib.sitemapNS) {
                    var res = this.oRequest.responseXML
                        .evaluate('//*[@id="'+sID+'"]',
                                  this.oRequest.responseXML,
                                  function() { return cand; },
                                  window.XPathResult.FIRST_ORDERED_NODE_TYPE,
                                  null).singleNodeValue;
                    if (res)
                        return res;
                }

                return res;
            },
            oRequest: oRequest,
            oSitemapURI: oSitemapURI,
            aActions: []
        };

        oEntry.aActions.push(function()
        {
            if (oRequest.readyState == 4) {
                oEntry.bOkay = false;
                if (oRequest.status == 200 || oRequest.status == 0) {
                    //self.lib.dump('File finally loaded');
                    oEntry.bOkay = true;
                } else if (oRequest.status == 400) {
                    self.lib.dump('Sitemap <' + oSitemapURI +
                           '>: Bad request.');
                } else if (oRequest.status == 401) {
                    self.lib.dump('Sitemap <' + oSitemapURI +
                           '>: Unauthorized access.');
                } else if (oRequest.status == 403) {
                    self.lib.dump('Sitemap <' + oSitemapURI +
                           '>: Forbidden.');
                } else if (oRequest.status == 404) {
                    self.lib.dump('Sitemap <' + oSitemapURI +
                           '>: File not found.');
                } else if (oRequest.status == 500) {
                    self.lib.dump('Sitemap <' + oSitemapURI +
                           '>: Internal server error.');
                } else if (oRequest.status == 503) {
                    self.lib.dump('Sitemap <' + oSitemapURI +
                           '>: Service unavailable.');
                } else {
                    self.lib.dump('Other error ' + oRequest.status + '.');
                }
            }
        });

        // ...then perform the caller's requested action.
        oEntry.aActions.push(fOnCompletion);

        // Arrange to perform all scheduled actions when the document
        // is ready.
        oRequest.onload = function()
        {
            for each(fAction in oEntry.aActions) fAction();
        };

        // Cache the request so that future requests for the same
        // document can simply attach themselves to this one.
        this.aRequestCache[oSitemapURI.spec] = oEntry;

        oRequest.open('GET', oSitemapURI.spec, true);
        //Optionally bypass the browser cache.
        if (this.bSkipCache) oRequest.channel.loadFlags |= window.Components.interfaces.nsIRequest.LOAD_BYPASS_CACHE;
        oRequest.send(null);
    },
};

StandardSitemapService.prototype.Sitemap.prototype.VariantData.prototype = {
    lib : Application.storage.get("org.standard-sitemap.library", null),

    decodeElement : function(mSpecs, oElement, mDefaults) {
        // Make a copy of the defaults.
        var mResult = { };
        if (mDefaults)
            for (var key in mDefaults)
                mResult[key] = mDefaults[key];

        if (oElement) {
            // Extract resultant characteristics from this element.

            // TODO: For attributes with natural-language content,
            // make note of xml:lang currently in effect.

            var tmp;

            if (typeof(oElement.getAttribute) != "function")
                alert('oElement is : ' + oElement);

            // Get the image width and height.
            tmp = oElement.getAttribute('width');
            if (tmp) mResult['width'] = tmp;
            tmp = oElement.getAttribute('height');
            if (tmp) mResult['height'] = tmp;
            tmp = oElement.getAttribute('bgcolor');
            if (tmp) mResult['bgcolor'] = tmp;

            // TODO: Get the name from a child element, then only from the
            // attribute as a back-up.
            tmp = oElement.getAttribute('name');
            if (tmp) mResult['name'] = tmp;

            // TODO: Get the description from a child element, then only
            // from the attribute as a back-up.
            tmp = oElement.getAttribute('description');
            if (tmp) mResult['desc'] = tmp;

            // Get the array of referrer preferences.
            tmp = oElement.getAttribute('refer');
            if (tmp)
                mResult['refer'] = tmp.trim().split(/\s+/);
            else
                mResult['refer'] = [ ];

            // Get @data and @method too, so we can have localized query
            // strings.
            tmp = oElement.getAttribute('data');
            if (tmp) mResult['searchdata'] = tmp;
            tmp = oElement.getAttribute('method');
            if (tmp) mResult['searchmethod'] = tmp;

            // Get the address from @xlink:href, else from @url.
            tmp = this.lib.getURIAttrNS(oElement, this.lib.xlinkNS, 'href');
            if (!tmp)
                tmp = this.lib.getURIAttr(oElement, 'url');
            if (tmp) {
                mResult['addr'] = tmp;
                this.mAddrs[tmp.spec] = true;
            }


            // Extract selection criteria from this element.

            // Is a language specified for this element?
            tmp = null;
            for (var ancestor = oElement; !tmp && oElement;
                 ancestor = ancestor.parentNode) {
                if (typeof(ancestor.getAttribute) != "function")
                    break;
                tmp = ancestor.getAttribute('lang');
            }
            if (tmp) mResult['lang'] = tmp.trim();

            // Is a format specified for this element?
            tmp = oElement.getAttribute('type');
            if (tmp) mResult['format'] = tmp.trim();
        }

        // Break the language code up into parts, so we can store this
        // entry under less specific keys.
        var sLang = mResult['lang'];
        if (sLang) {
            var aLangParts = sLang.trim().toLowerCase().split(/[-_.]/);
        } else {
            var aLangParts = [];
        }
        mResult['lang_spec'] = aLangParts.length;
        mResult['format_spec'] = mResult['format'] ? 1 : 0;

        var sep = [ '', '-', '.' ];
        var langCand = '';
        for (var i in aLangParts) {
            langCand += sep[i];
            langCand += aLangParts[i].toLowerCase();
            this.aLangs[langCand] = true;
        }

        // Populate the map.
        for (var iWithFormat in [ false, true ]) {
            var bWithFormat = [false, true][iWithFormat];
            if (bWithFormat && !mResult['format'])
                continue;
            for (var depth = 0; depth <= aLangParts.length; depth++) {
                var key = bWithFormat ? mResult['format'] : '/';
                for (var i = 0; i < depth; i++)
                    key += '/' + aLangParts[i];

                // Try to insert this mapping, but is there an
                // existing value?
                var old = this.mEntries[key];
                if (old) {
                    // If any of the key dimensions in the existing
                    // entry is less specific than those in the new
                    // entry, we can't overwrite the existing entry.

                    var bIsBetterDim = false;

                    for (var sDim in { 'format' : 0, 'lang' : 0 }) {
                        // Compare the old dimension with the new.
                        var iDiff = old[sDim + '_spec'];
                        iDiff -= mResult[sDim + '_spec'];
                        if (iDiff < 0) {
                            // There's no way to replace this entry.
                            bIsBetterDim = false;
                            break;
                        }
                        if (iDiff > 0)
                            // Provided there's no other dimension
                            // that fails outright, this one would be
                            // an improvement.
                            bIsBetterDim = true;
                    }
                    if (!bIsBetterDim) {
                        // Copy any fields that the existing entry
                        // doesn't have.
                        for (var f in mResult)
                            if (!(f in old))
                                old[f] = mResult[f];
                        continue;
                    }
                }

                // Okay, insert now.
                this.mEntries[key] = mResult;
            }
        }

        if (oElement) {
            // Check if there are any <variant>s to this element.
            var iChild, iLen = oElement.childNodes.length;
            for (iChild = 0; iChild < iLen; iChild++) {
                var oChild = oElement.childNodes[iChild];
                if (oChild.nodeType != oChild.ELEMENT_NODE)
                    continue;
                if (oChild.tagName != 'variant')
                    continue;
                if (!this.lib.issmns(oChild.namespaceURI))
                    continue;
                this.decodeElement(mSpecs, oChild, mResult);
            }
        }
    },

    addEntry : function(mSpecs, key, value, langSpec) {
        var oldValue = this.mEntries[key];
        if (oldValue && mSpecs[key] < langSpec)
            return;
        this.mEntries[key] = value;
    },

    // Search through this data for an entry matching the language and
    // format.  Null arguments mean 'don't care'.  Return a map of
    // 'name', 'desc' and 'addr'.  'name' and 'desc' will typically be
    // strings, but future implementations might allow some marked up
    // content in the form of DOM elements.  'addr' should normally be
    // an nsIURI object.
    get : function(aLang, aFormat) {
        if (!aLang) aLang = [ null ];
        if (!aFormat) aFormat = [ null ];

        for (var iLang in aLang) {
            var sLang = aLang[iLang];
            sLang = sLang
                ? '/' + sLang.toLowerCase().replace("[-_.]", "/")
                : "";

            for (var iFormat in aFormat) {
                var sFormat = aFormat[iFormat];

                var key = sFormat ? sFormat : '/';
                key += sLang;

                var result = this.mEntries[key];
                if (result)
                    return result;
            }
        }

        return null;
    },

    // Get a map(string->boolean) of language strings that could be
    // presented in the sitemap's language selection.
    getLangs : function() {
        return this.aLangs;
    },

    /* Convenience methods */

    getField : function(sField, aArgs) {
        var result = this.get.apply(this, aArgs);
        if (!result) return null;
        if (sField in result) return result[sField];
        return null;
    },

    getName : function() {
        return this.getField('name', arguments);
    },

    getDescription : function() {
        return this.getField('desc', arguments);
    },

    getAddress : function() {
        return this.getField('addr', arguments);
    },

    getQuery : function() {
        return this.getField('searchdata', arguments);
    },

    getMethod : function() {
        return this.getField('searchmethod', arguments);
    }
};

/*Node represents an individual node from the standard sitemap xml file*/
StandardSitemapService.prototype.Sitemap.prototype.Node.prototype =
{
    lib : Application.storage.get("org.standard-sitemap.library", null),

//     oParentNode: null, //Parent node.
    aChildNodes: null, //Array containing child Nodes
    oVariants : null, // Holds negotiable content
    sID: "", //Identifier of this node to be able to match a UI element to it.
    sRole: "", //Contains the item's role.
    sRelation: "", //Contains the item's relation.
    iPriority: 0, //Contains the item's priority as a CSS font-weight value 100-900.
    bIsGroup: false, //Is the item a group or item node.
//     oSourceElement: null, // For detecting recursive loops
    mNext : null,
    mPrev : null,
    mParent : null,
    sStateId : null,

    // The address of the XML file from which this element is taken
    oSitemapURI : null,

    // TODO: Change mPrev/mNext so that the value is not merely the
    // sibling node, but a pair including the sibling and the
    // effective parent.  This can be used as a back-up if the direct
    // parent is unable to provide a name or a URI.

    getSitemapURI : function() { return this.oSitemapURI; },

    /*Initialize a node*/
    Initialize: function()
    {
        var self = this;
        var hashFunc = function(obj) {
            if ("standardSitemapHashCode" in obj)
                return obj.standardSitemapHashCode;
            return obj.standardSitemapHashCode = self.nextHashCode++;
        };

        this.mNext = this.lib.newMap(null, hashFunc);
        this.mPrev = this.lib.newMap(null, hashFunc);
        this.mParent = this.lib.newMap(null, hashFunc);
        this.aChildNodes = new Array();
        this.oVariants = null;
        this.sID = "ssmn_" + this.lib.GenerateID();
    },
    
    /*Set some common properties*/
    SetProperties: function(oVariants, sRole, sPriority, sRelation, bIsGroup)
    {
        //ToDo: If the node would for some reason contain a uri and a name, but also have variant children containinf the multilingual attributes
        //      we might have a problem here. Cause then the <item>'s attributes would be stored as the first language, and the <variant>'s as next.
        this.oVariants = oVariants;
        this.sRole = sRole;
        this.sRelation = sRelation;
        this.iPriority = this.Priority2FontWeight(sPriority);
        this.bIsGroup = bIsGroup;
    },
    
    /*Reset all properties*/
    Reset: function()
    {
//         this.oParentNode = null;
        this.aChildNodes = null;
        this.oVariants = null;
        this.sID = "";
        this.sRole = "";
        this.sRelation = "";
        this.ResetChildNodes();
        this.iPriority = 0;
        this.bIsGroup = false;
        this.mNext.clear();
        this.mPrev.clear();
        this.mParent.clear();
    },

    /*Reset all child nodes*/
    ResetChildNodes: function()
    {
        if (this.aChildNodes)
        {
            var iIndex, iLen = this.aChildNodes.length;
            for (iIndex = 0; iIndex < iLen; iIndex ++) this.aChildNodes[iIndex].Reset();
        }
    },

    /* Get the content of this node negotiated against language/format
    preferences. */
    GetContent : function() {
        return this.oVariants.get.apply(this.oVariants, arguments);
    },

    /* Check whether the XML element is already represented by this
    node or any of its ancestors. */
//     IsNodeAccountedFor: function(oElement) {
//         // The 'root' node has no associated source element.
//         if (!this.oSourceElement)
//             return false;
//         if (this.oSourceElement == oElement)
//             return true;
//         if (!this.oParentNode)
//             return false;
//         return this.oParentNode.IsNodeAccountedFor(oElement);
//     },

    /*Add a child node, returns the newly created node*/
    AddChildNode: function(oElement)
    {
        // Check for loops.
//         if (this.IsNodeAccountedFor(oElement))
//             return null;

        // TODO: Having to pass in the constructor you're using seems
        // to be an ugly hack, but this.constructor just returns
        // Object, not the Node's constructor.  Is there a better way?
        var oNode = new (this.creator)(this.creator);
        this.aChildNodes.push(oNode);
//         oNode.oParentNode = this;
        oNode.mParent.put(this, true);
//         oNode.oSourceElement = oElement;

        return oNode;
    },

    /* Gets the name for the given language preference. */
    GetName: function(aLanguages)
    {
        if (this.oVariants && typeof(this.oVariants.getName) == "function")
            return this.oVariants.getName(aLanguages);
        return null;
    },

    /* Gets the description for the given language preference. */
    GetDescription: function(aLanguages)
    {
        if (this.oVariants &&
            typeof(this.oVariants.getDescription) == "function")
            return this.oVariants.getDescription(aLanguages);
    },

    /* This is just meant to be a fail-safe debugging tool.  If GetURI
    returns null, this returns null.  Otherwise, it converts the URI
    to a string. */
    GetURISpec : function(aLanguages)
    {
        var r = this.GetURI(aLanguages);
        if (r) return r.spec;
    },

    /* Gets the uri for the given language preference. */
    GetURI: function(aLanguages)
    {
        if (this.oVariants && typeof(this.oVariants.getAddress) == "function")
            return this.oVariants.getAddress(aLanguages);
    },
    
    /*Converts a sitemap priority (0-1) to a valid CSS font-weight value 100-900*/
    Priority2FontWeight: function(sPriority)
    {
        if (!sPriority)
            sPriority = 0.5;
        // This is so complicated because priority=0.5 (=default) has to be mapped to font-weight=400 (=default):
        if (sPriority <= 0.5)
                var iItemPriority = Math.round((sPriority*600+100)/100)*100; // f=p*600+100 converts from 0-0.5 to 100-400
        else
                var iItemPriority = Math.round((sPriority*1000-100)/100)*100; // f=p*1000-100 converts from 0.5-1 to 400-900
        // Results of the mapping:
        // sPriority    : 0.0   0.1   0.2   0.3   0.4   0.5   0.6   0.7   0.8   0.9   1.0
        // iItemPriority: 100   200   200   300   300   400   500   600   700   800   900
        return iItemPriority;    
    },

    /*Boolean check if the node has childnodes*/
    get HasChildren()
    {
        return (this.aChildNodes.length > 0);
    },
    
    /*Returns the 1st child element*/
    get FirstChild()
    {
        if (this.HasChildren) return this.aChildNodes[0];
        return null;
    },

    /*Returns the last child element*/
    get LastChild()
    {
        if (this.HasChildren) return this.aChildNodes[this.aChildNodes.length - 1];
        return null;
    }    
};

/*Logo holds the prcoessed sitemap logo + url*/
StandardSitemapService.prototype.Sitemap.prototype.Logo.prototype = {
    lib : Application.storage.get("org.standard-sitemap.library", null),

    oImageVariants : null,
    oLinkVariants : null,
    
    /*Initialize the logo*/
    SetProperties: function(oImageVariants, oLinkVariants)
    {
        this.oImageVariants = oImageVariants;
        this.oLinkVariants = oLinkVariants;
    },
    
    /*Reset all properties*/
    Reset: function()
    {
        this.oImageVariants = null;
        this.oLinkVariants = null;
    }
};


Application.storage.set("org.standard-sitemap.service",
                        new StandardSitemapService());
